snpden -Kmax 5 -scr eclsrc.out -out eclout.out ecl.dat > eclplt.out
compare eclsrc.out eclsrc.bak
compare eclout.out eclout.bak
compare eclplt.out eclplt.bak
snpden -Kmax 5 -bic ecl.dat > eclbic.out
compare eclbic.out eclbic.bak
snpden -Kmax 5 -aic ecl.dat > eclaic.out
compare eclaic.out eclaic.bak
snpden -Kmax 3 -p ecl.dat > eclp.out
compare eclp.out eclp.bak
