snpden -Kmax 5 -scr eclsrc.out -out eclout.out ecl.dat >eclplt.out 2>err1.out
diff eclsrc.out eclsrc.bak
diff eclout.out eclout.bak
diff eclplt.out eclplt.bak
diff err1.out err1.bak
snpden -Kmax 5 -bic ecl.dat > eclbic.out 2>err2.out
diff eclbic.out eclbic.bak
diff err2.out err2.bak
snpden -Kmax 5 -aic ecl.dat > eclaic.out 2>err3.out
diff eclaic.out eclaic.bak
diff err3.out err3.bak
snpden -Kmax 3 -p ecl.dat > eclp.out 2>err4.out
diff eclp.out eclp.bak
diff err4.out err4.bak
